import time
import torch
import torch.nn as nn

from tqdm import tqdm
from typing import Optional, Tuple

def soft_thresh(x: torch.Tensor,
                rho: float) -> torch.Tensor:
    """
    Soft thresholding operator

    Parameters:
    -----------
    x : torch.Tensor
        input tensor
    rho : float
        threshold value

    Returns:
    --------
    x_thresh : torch.Tensor
        thresholded tensor
    """
    return torch.exp(1j * torch.angle(x)) * torch.max(torch.abs(x) - rho, torch.zeros(x.shape, device=x.device))

def power_method_operator(A: callable,
                          x0: torch.Tensor,
                          num_iter: Optional[int] = 15,
                          verbose: Optional[bool] = True) -> Tuple[torch.Tensor, float]:
    """
    Uses power method to find largest eigenvalue and corresponding eigenvector

    Parameters:
    -----------
    A : callable
        linear operator
    vec_init : torch.Tensor
        initial guess of eigenvector with shape (*vec_shape)
    num_iter : int
        number of iterations to run power method
    verbose : bool
        toggles progress bar
    
    Returns:
    --------
    eigen_vec : torch.Tensor
        eigenvector with shape (*vec_shape)
    eigen_val : float
        eigenvalue
    """
    
    for _ in tqdm(range(num_iter), 'Max Eigenvalue', disable=not verbose):
        
        z = A(x0)
        ll = torch.norm(z)
        x0 = z / ll
    
    if verbose:
        print(f'Max Eigenvalue = {ll}')
    
    return x0, ll.item()

def FISTA(AHA: callable, 
          proxg: callable,
          AHb: torch.Tensor,  
          num_iters: Optional[int] = 20,
          ptol_exit: Optional[float] = 0.5,
          verbose: Optional[bool] = True) -> torch.Tensor:
    """
    Solves ||Ax - b||_2^2 + lamda ||Gx||_1, where G is a linear function.

    Parameters
    ----------
    AHA : callable
        The gram or normal operator of A
    proxg : callable
        The proximial operator for ||Gx||_1
    AHb : torch.tensor
        The A hermitian transpose times b
    num_iters : int
        Number of iterations
    ptol_exit : float
        percent tolerance exit condition
    verbose : bool
        toggles print statements

    Returns
    ---------
    x : torch.tensor
        Reconstructed tensor
    """

    # Start at AHb
    x = AHb.clone()
    z = x.clone()
    
    if num_iters <= 0:
        return x

    # Benchmarking
    t_gr = 0.0
    t_prox = 0.0

    for k in tqdm(range(0, num_iters), 'FISTA Iterations', disable=not verbose):

        x_old = x.clone()
        x     = z.clone()
        
        start = time.perf_counter()
        gr    = AHA(x) - AHb
        end = time.perf_counter()
        t_gr += end - start
        start = time.perf_counter()
        x     = proxg(x - gr)
        end = time.perf_counter()
        t_prox += end - start
        if k == 0:
            z = x
        else:
            step  = k/(k + 3)
            z     = x + step * (x - x_old)
        ptol = 100 * torch.norm(x_old - x)/torch.norm(x)
        if ptol < ptol_exit:
            if verbose:
                print(f'Tolerance reached after {k+1} iterations, exiting FISTA')
            break

    if verbose:
        print(f'Gradient Took {t_gr:.3f}(s), Prox Took {t_prox:.3f}(s)')
    return x

def ADMM(AHA: callable,
         D: callable,
         DH: callable,
         AHb: torch.Tensor,
         rho: float,
         lamda: float,
         lamda_l2: Optional[float] = 0.0,
         n_cg_iter: Optional[int] = 50,
         num_iters: Optional[int] = 100,
         DHD: Optional[callable] = None,
         denoiser: Optional[callable] = None,
         verbose: Optional[bool] = True) -> torch.Tensor:
    """
    Alternating Direction Method of Multipliers (ADMM) for solving:
    argmin_x 1/2||Ax - b||^2 + lamda ||z||^1
    s.t. Dx = z

    Parameters:
    -----------
    AHA : nn.Module 
        Linear operator representing the gram/normal operator of A
    D : callable
        Linear operator representing the forward operator of D
    DH : callable
        Linear operator representing the hermitian transpose of D
    AHb : torch.tensor
        The A hermitian transpose times b
    rho : float
        ADMM parameter enforcing constraint
    lamda : float
        Regularization strength
    num_iters : int
        Max number of iterations.
    DHD : nn.Module
        Linear operator representing the gram/normal operator of D,
        only given if more efficeint than calling DH(D(.))
    denoiser : callable
        General denoiser to replace L1 norm
    verbose : bool
        toggles print statements
    """
    
    # Define relevant linops
    if DHD is None:
        DHD = lambda x : DH(D(x))
    AHA_tilde = lambda x : AHA(x) + rho * DHD(x)

    # Introduce slack variables z and u
    x = torch.zeros_like(AHb)
    z = D(x)
    u = D(x)
    
    for i in tqdm(range(num_iters), 'ADMM Iterations', disable=not verbose):
        
        # X-update
        x = conjugate_gradient(AHA_tilde, AHb + rho * DH(z - u), num_iters=n_cg_iter, lamda_l2=lamda_l2, verbose=False)

        # Z-update
        v = D(x) + u
        if denoiser is None:
            z = soft_thresh(v, lamda / rho)
        else:
            # Denoise
            z.real = rho * v.real + lamda * denoiser(v[None, None, ...].real).squeeze()
            z.imag = rho * v.imag + lamda * denoiser(v[None, None, ...].imag).squeeze()
            z = z / (rho + lamda)

        # U-update
        u = v - z
    
    return x

def gradient_descent(AHA: callable,
                     AHb: torch.Tensor,
                     lr: float,
                     num_iters: Optional[int] = 100,
                     lamda_l2: Optional[float] = 0.0,
                     tolerance: Optional[float] = 1e-8,
                     return_resids: Optional[bool] = False,
                     verbose: Optional[bool] = True) -> torch.Tensor:
    """
    Parameters:
    -----------
    AHA : nn.Module 
        Linear operator representing the gram/normal operator of A
    AHb : torch.tensor
        The A hermitian transpose times b
    lr : float
        learning rate
    num_iters : int 
        Max number of iterations.
    lamda_l2 : float
        Replaces AHA with AHA + lamda_l2 * I
    tolerance : float 
        Used as stopping criteria
    return_resids : bool
        toggles return of residuals
    verbose : bool
        toggles print statements
    
    Returns:
    ---------
    x : torch.tensor <complex>
        least squares estimate of x
    """
    x = AHb
    AHA_wrapper = lambda x : AHA(x) + lamda_l2 * x
    resids = []
    for i in tqdm(range(num_iters), 'GD Iterations', disable=not verbose):
        grad = AHA_wrapper(x) - AHb
        x = x - lr * grad
        nrm = torch.norm(grad)
        resids.append(nrm.item())
        if nrm < tolerance:
            break
    if return_resids:
        return resids, x
    else:
        return x

def conjugate_gradient(AHA: callable, 
                       AHb: torch.Tensor, 
                       P: Optional[nn.Module] = None,
                       num_iters: Optional[int] = 10, 
                       lamda_l2: Optional[float] = 0.0,
                       tolerance: Optional[float] = 1e-8,
                       return_resids: Optional[bool] = False,
                       verbose=True) -> torch.Tensor:
    """Conjugate gradient for complex numbers. The output is also complex.
    Solve for argmin ||Ax - b||^2. Inspired by sigpy!
    
    Parameters:
    -----------
    AHA : nn.Module 
        Linear operator representing the gram/normal operator of A
    AHb : torch.tensor
        The A hermitian transpose times b
    P : nn.Module
        Preconditioner 
    num_iters : int 
        Max number of iterations.
    lamda_l2 : float
        Replaces AHA with AHA + lamda_l2 * I
    tolerance : float 
        Used as stopping criteria
    return_resids : bool
        toggles return of residuals
    verbose : bool
        toggles print statements
    
    Returns:
    ---------
    x : torch.tensor <complex>
        least squares estimate of x, same shape as x0 if provided    
    """

    # Default preconditioner is identity matrix
    if P is None:
        P = lambda x : x
    
    # Tikonov regularization
    AHA_wrapper = lambda x : AHA(x) + lamda_l2 * x

    # Start at AHb
    x0 = AHb.clone()
    if num_iters == 0:
        return x0

    # Define iterative vars
    r = AHb - AHA_wrapper(x0)
    z = P(r)
    p = z.clone()

    resids = []

    # Main loop
    for i in tqdm(range(num_iters), 'CG Iterations', disable=not verbose):

        # Apply model
        Ap = AHA_wrapper(p)
        pAp = torch.real(torch.sum(p.conj() * Ap)).item()

        # Update x
        # assert pAp > 0, 'A is not Semi-Definite'
        rz = torch.real(torch.sum(r.conj() * z))
        alpha = rz / pAp
        x0 = x0 + alpha * p

        # Update r
        r = r - alpha * Ap
        rnrm = torch.norm(r)
        resids.append(rnrm.item())
        if rnrm < tolerance:
            break

        # Update z
        z = P(r)

        # Update p
        beta = torch.real(torch.sum(r.conj() * z)) / rz
        p = z + beta * p
    
    if return_resids:
        return resids, x0
    else:
        return x0