import torch

from typing import Optional, Tuple
from wearable_mri.mag_sim import loop_coil

def gen_pts_on_circle(n_pts: int, 
                      r: float,
                      center: torch.Tensor,
                      normal: torch.Tensor):
    """
    Generate n_pts on a circle with radius r about 
    some center point and normal vector.

    Parameters:
    -----------
    n_pts : int
        Number of points to generate
    r : float
        Radius of the circle
    center : torch.Tensor
        Center of the circle with shape (3,)
    normal : torch.Tensor
        Normal vector to the circle with shape (3,)

    Returns:
    --------
    pts : torch.Tensor
        Points on the circle with shape (n_pts, 3)
    """

    # Make x hat and y hat vectors
    normal = normal / torch.norm(normal)
    x_hat = torch.tensor([normal[1] + normal[2],
                            normal[2] - normal[0],
                            -normal[0] - normal[1]], 
                            device=center.device,
                            dtype=center.dtype)
    x_hat = x_hat / torch.norm(x_hat)
    y_hat = torch.cross(normal, x_hat)
    y_hat = y_hat / torch.norm(y_hat)
    assert normal @ x_hat < 1e-6
    assert normal @ y_hat < 1e-6
    assert x_hat @ y_hat < 1e-6

    # Generate points
    thetas = torch.linspace(0, 2*torch.pi, n_pts)[:, None]
    pts = r * (torch.cos(thetas) * x_hat + torch.sin(thetas) * y_hat) + center
    return pts

def visualize_coils(loop_coils: list[loop_coil],
                    colors: Optional[list[str]] = None):
    """
    Visualize the coil responses

    Parameters:
    -----------
    loop_coils : list[loop_coil]
        List of loop_coil objects
    colors : list[str]
        List of colors to use for each coil
    """

    # Consts
    n_coils = len(loop_coils)

    # Create figure
    import matplotlib.pyplot as plt
    fig = plt.figure(figsize=(14,7))
    ax = fig.add_subplot(projection='3d')

    
    # Plot
    for i in range(n_coils):
        
        # Get coil info
        center = loop_coils[i].center.cpu()
        normal = loop_coils[i].normal.cpu()
        radius = loop_coils[i].diameter / 2
        nturns = loop_coils[i].nturns

        if colors is not None:
            color = colors[i]
        else:
            color = 'black'

        # Gen points on circle
        pts = gen_pts_on_circle(100, radius, center, normal)

        # Plot
        ax.plot(pts[:, 0], pts[:, 1], pts[:, 2], 
                label=f'Coil {i+1}', 
                linewidth=round(nturns/100),
                color=color,
                alpha=0.4)

def pre_polarizing_coils(pts: torch.Tensor,
                         spacing: float,
                         r_coil: float,
                         nturns: Optional[int] = 100) -> Tuple[torch.Tensor, list[loop_coil]]:
    """
    Generate a strong polarizing field from 2 coils with normal vectos in Z direction 
    with some spacing in the z-direction

    Parameters:
    -----------
    pts : torch.Tensor
        Points of interest with shape (..., 3) (assuming last dim is zeros)
    spacing : float
        Spacing between the coils in the z-direction [m]
    r_coil : float
        Radius of the coil [m]
    nturns : int
        Number of turns in the coil
    
    Returns:
    --------
    coil_responses : torch.Tensor
        responses per coil with shape (2, ..., 3)
    loop_coils : list[loop_coil]
        List of loop_coil objects
    """

    # Consts
    device = pts.device
    dtype = pts.dtype

    # coil centers
    centers = torch.tensor([
        [0, 0, -spacing/2],
        [0, 0, spacing/2]
    ], device=device, dtype=dtype)

    # normal vectors
    normals = torch.tensor([
        [0, 0, 1],
        [0, 0, 1]
    ], device=device, dtype=dtype)

    # Make coils
    lc1 = loop_coil(
        normal=normals[0],
        center=centers[0],
        diameter=r_coil * 2,
        current=1,
        nturns=nturns
    )
    lc2 = loop_coil(
        normal=normals[1],
        center=centers[1],
        diameter=r_coil * 2,
        current=1,
        nturns=nturns
    )
    loop_coils = [lc1, lc2]
    responses = torch.stack((lc1.calc_field(pts), lc2.calc_field(pts)), dim=0)

    return responses, loop_coils

def surface_coils_2d(n_coils: int,
                     pts: torch.Tensor,
                     r_circle: Optional[float] = 0.1,
                     r_coil: Optional[float] = None,
                     nturns: Optional[int] = 100) -> Tuple[torch.Tensor, list[loop_coil]]:
    """
    Places n_coils around surface of a circle in the xy-plane, 
    and returns the 1A response of each coil

    Parameters:
    -----------
    n_coils : int
        Number of coils to place
    pts : torch.Tensor
        Points of interest with shape (..., 3) (assuming last dim is zeros)
    r_circle : float
        Radius of the circle [m]
    r_coil : float
        Radius of the coil [m]
    nturns : int
        Number of turns in the coil

    Returns:
    --------
    coil_responses : torch.Tensor
        List of coil responses with shape (n_coils, ..., 3)
    loop_coils : list[loop_coil]
        List of loop_coil objects
    """

    # Consts
    device = pts.device
    dtype = pts.dtype

    if r_coil is None:
        r_coil = torch.pi * r_circle / n_coils

    # Generate points on circle
    thetas = torch.linspace(0, 2*torch.pi, n_coils + 1, device=device, dtype=dtype)
    thetas = (thetas[:-1] + thetas[1:]) / 2

    # coil centers
    centers = torch.stack([
        r_circle * torch.cos(thetas),
        r_circle * torch.sin(thetas),
        torch.zeros(n_coils, device=device, dtype=dtype)
    ], dim=1)

    # normal vectors
    normals = -centers / torch.norm(centers, dim=1, keepdim=True)

    # Make coils
    loop_coils = []
    responses = []
    for i in range(n_coils):
        lc = loop_coil(
            normal=normals[i],
            center=centers[i],
            diameter=r_coil * 2,
            current=1,
            nturns=nturns)
        
        responses.append(lc.calc_field(pts))
        loop_coils.append(lc)
    responses = torch.stack(responses, dim=0)

    return responses, loop_coils
