import torch
import torch.nn as nn

from typing import Optional
from tqdm import tqdm
from einops import einsum, rearrange

class encoding_linop(nn.Module):

    def __init__(self,
                 b0_steady_state: torch.Tensor,
                 coil_responses: torch.Tensor,
                 currents_imaging: torch.Tensor,
                 times: torch.Tensor,
                 coil_batch_size: Optional[int] = None,
                 shot_batch_size: Optional[int] = None,
                 voxl_batch_size: Optional[int] = None,
                 verbose: Optional[bool] = True):
        """
        Parameters:
        -----------
        b0_steady_state : torch.Tensor
            steady state field with shape (*im_size, 3)
        coil_responses : torch.Tensor
            List response field [T] from unit current with shape (n_coils, *im_size, 3)
        currents_imaging : torch.Tensor
            Currents for imaging with shape (n_shots, n_coils)
        times : torch.Tensor
            Time [s] points to image at with shape (n_read,)
            assumed to start at echo time!
        coil_batch_size : int
            Number of coils to batch 
        verbose : bool
            Print progress bar
        """
        super(encoding_linop, self).__init__()

        # Constants
        self.n_coils = coil_responses.shape[0]
        self.n_read = times.shape[0]
        self.n_shots = currents_imaging.shape[0]
        self.im_size = coil_responses.shape[1:-1]
        if coil_batch_size is None:
            coil_batch_size = self.n_coils
        if shot_batch_size is None:
            shot_batch_size = self.n_shots
        if voxl_batch_size is None:
            voxl_batch_size = torch.prod(torch.tensor(self.im_size)).item()
        self.shot_batch_size = shot_batch_size
        self.coil_batch_size = coil_batch_size
        self.voxl_batch_size = voxl_batch_size
        self.verbose = verbose
        
        # Make steady state field
        eps = 1e-6
        self.b0_ss_mag = torch.norm(b0_steady_state, dim=-1)
        self.b0_ss_dir = b0_steady_state / (self.b0_ss_mag[..., None] + eps)
        
        # Store coil responses and times 
        self.coil_responses = coil_responses
        self.times = times
        self.AHA = None

        # Store imaging currents
        self.currents_imaging = currents_imaging
    
    def get_encoding_functions(self, 
                               shot_slice: Optional[slice] = slice(None)):
        """
        Returns the encoding functions in the forward model 

        Parameters:
        -----------
        shot_slice : slice
            Slice the shot dimension
        
        Returns:
        --------
        B0_encoding : torch.Tensor
            Weighting from precession with shape (n_read, n_shots, *im_size)
        B1_encoding : torch.Tensor
            Receive coil weighting with shape (n_coils, n_shots, *im_size)
        """

        # Constants
        coil_responses = self.coil_responses
        currents_imaging = self.currents_imaging[shot_slice]
        b0_ss_dir = self.b0_ss_dir
        GAMMA_BAR = 42.58e6 # Hz/T
        eps = 1e-6
        nvox = torch.prod(torch.tensor(self.im_size)).item()

        b0_im = einsum(coil_responses, currents_imaging, 'nc ... three, npe nc -> npe ... three')
        b0_im_mag = torch.norm(b0_im, dim=-1)
        b0_im_dir = b0_im / (b0_im_mag[..., None] + eps)

        # Precession frequency
        omega = 2 * torch.pi * GAMMA_BAR * b0_im_mag # npe ...

        # x_hat corresponds to real, y_hat to imaginary
        x_hat = b0_ss_dir[None,] - torch.sum(b0_im_dir * b0_ss_dir[None,], dim=-1, keepdim=True) * b0_im_dir # npe ... 3
        x_hat /= (torch.norm(x_hat, dim=-1, keepdim=True) + eps) # npe ... 3
        y_hat = torch.linalg.cross(x_hat, b0_im_dir, dim=-1)
        y_hat /= (torch.norm(y_hat, dim=-1, keepdim=True) + eps) # npe ... 3

        # Make analytical coil response
        B1 = einsum(coil_responses, x_hat, 'nc ... three, npe ... three -> nc npe ...') + 1j * \
             einsum(coil_responses, y_hat, 'nc ... three, npe ... three -> nc npe ...')
            
        # Get d/dt{m(r, t)} TODO demodulate?
        phz = torch.exp(-1j * einsum(self.times, omega, 'nro, npe ... -> nro npe ...'))
        mag = omega * self.b0_ss_mag
        
        B0_encoding = phz * mag / nvox
        B1_encoding = B1
        
        return B0_encoding, B1_encoding

    def forward(self,
                m: torch.Tensor) -> torch.Tensor:
        """
        Applies the linear system model to m given the imaging currents

        Parameters:
        -----------
        m : torch.Tensor
            Magnetization at steady state with shape (im_size)
            Should just be T2 and PD weighted image
        
        Returns:
        --------
        b : torch.Tensor
            received data (complex) with shape (n_coils, n_read, n_shots)
        """
        # Constants
        n_coils = self.n_coils
        n_shots = self.n_shots
        n_read = self.n_read
        coil_batch_size = self.coil_batch_size
        verbose = self.verbose

        # Loop over all imaging currents and return all imaging matrices
        b = torch.zeros((n_coils, n_read, n_shots), dtype=torch.complex64, device=m.device)
        for s1 in range(0, n_shots, self.shot_batch_size):
            s2 = min(s1 + self.shot_batch_size, n_shots)

            # Get encoding functions
            B0_enc, B1_enc = self.get_encoding_functions(shot_slice=slice(s1, s2))

            # Batch over coils
            for c1 in tqdm(range(0, n_coils, coil_batch_size), 'Forward Model', disable=not verbose):
                c2 = min(c1 + coil_batch_size, n_coils)

                # Integrate 
                enc = B0_enc[None, :, ...] * B1_enc[c1:c2, None, ...]
                b[c1:c2, :, s1:s2] = einsum(enc, m, 'nc nro npe ..., ... -> nc nro npe')  

        return b
        
    def adjoint(self,
                b: torch.Tensor) -> torch.Tensor:
        """
        Applies the adjoint of the linear system model to b given the imaging currents

        Parameters:
        -----------
        b : torch.Tensor
            received data (complex) with shape (n_coils, n_read, n_shots)
        """

        # Constants
        im_size = self.im_size
        n_coils = self.n_coils
        n_shots = self.n_shots
        coil_batch_size = self.coil_batch_size
        verbose = self.verbose

        # Loop over all imaging currents and return all imaging matrices
        m = torch.zeros(im_size, dtype=torch.complex64, device=b.device)
        for s1 in range(0, n_shots, self.shot_batch_size):
            s2 = min(s1 + self.shot_batch_size, n_shots)

            # Get encoding functions
            B0_enc, B1_enc = self.get_encoding_functions(shot_slice=slice(s1, s2))
            
            # Batch over coils            
            for c1 in tqdm(range(0, n_coils, coil_batch_size), 'Forward Model', disable=not verbose):
                c2 = min(c1 + coil_batch_size, n_coils)

                # Integrate 
                enc = B0_enc[None, :, ...] * B1_enc[c1:c2, None, ...]
                m += einsum(enc.conj(), b[c1:c2, :, s1:s2], 'nc nro npe ..., nc nro npe -> ...')

        return m

    def calc_AHA_matrix(self,
                        currents_imaging: torch.Tensor):
        """
        Calculates AHA( . ) = adjoint(forward( . )) matrix

        Parameters:
        -----------
        currents_imaging : torch.Tensor
            Currents for imaging with shape (n_shots, n_coils)

        Saves:
        --------
        self.AHA : torch.Tensor
            AHA matrix with shape (*im_size, *im_size)
        """

        # Constants
        im_size = self.im_size
        n_coils = self.n_coils
        n_shots = self.n_shots
        coil_batch_size = self.coil_batch_size
        shot_batch_size = self.shot_batch_size

        # Loop over all imaging currents and return all imaging matrices
        nvox = torch.prod(torch.tensor(im_size)).item()
        AHA = torch.zeros((nvox, nvox), dtype=torch.complex64, device=self.b0_ss_mag.device)
        for s1 in tqdm(range(0, n_shots, shot_batch_size), 'Building AHA Matrix'):
            s2 = min(s1 + shot_batch_size, n_shots)

            # Get encoding functions
            B0_enc, B1_enc = self.get_encoding_functions(shot_slice=slice(s1,s2))
            
            # Batch over coils            
            for c1 in range(0, n_coils, coil_batch_size):
                c2 = min(c1 + coil_batch_size, n_coils)

                # Integrate 
                enc = B0_enc[None, :, ...] * B1_enc[c1:c2, None, ...] # nc nro npe ...
                enc = rearrange(enc, 'nc nro npe ... -> (nc nro npe) (...)') # T R
                for r1 in range(0, nvox, self.voxl_batch_size):
                    r2 = min(r1 + self.voxl_batch_size, nvox)
                    AHA[r1:r2] += enc.H[r1:r2] @ enc
        self.AHA = AHA

    def normal(self,
               m: torch.Tensor):
        """
        Applies the linear system model's normal operator to m given the imaging currents

        Parameters:
        -----------
        m : torch.Tensor
            Magnetization at steady state with shape (im_size)
            Should just be T2 and PD weighted image
        
        Returns:
        --------
        m_hat : torch.Tensor
            image after AHA applied with shape (im_size)
        """
        if self.AHA is None:
            return self.adjoint(self.forward(m))
        else:
            m_flt = rearrange(m, 'nx ny -> (nx ny)')
            m_hat = torch.zeros_like(m_flt)
            for r1 in range(0, m_flt.shape[0], self.voxl_batch_size):
                r2 = min(r1 + self.voxl_batch_size, m_flt.shape[0])
                m_hat[r1:r2] = self.AHA[r1:r2] @ m_flt
            return rearrange(m_hat, '(nx ny) -> nx ny', nx=self.im_size[0], ny=self.im_size[1])
    
    def visualize_field(self,
                        currents_imaging: torch.Tensor):
        """
        Visualizes the field over the phantom
        """

        import matplotlib.pyplot as plt
        bfield = einsum(self.coil_responses, currents_imaging, 'nc ... three, nc -> ... three')
        bfield = torch.linalg.norm(bfield * 1e3, dim=-1).cpu()
        figsize=(14, 7)
        fig, axs = plt.subplots(nrows=1, ncols=3, figsize=figsize)
        vmax = 1.0
        vmin = -vmax
        for j in range(3):
            axs[j].set_yticklabels([])
            axs[j].set_xticklabels([])
            axs[j].set_xticks([])
            axs[j].set_yticks([])
        im = axs[0].imshow(bfield, vmin=vmin, vmax=vmax, cmap='jet')
        axs[0].set_title('Field X [mT]')
        fig.colorbar(im)
        # axs[1].imshow(bfield[..., 1], cmap='jet', vmin=vmin, vmax=vmax)
        # axs[1].set_title('Field Y [mT]')
        # axs[2].imshow(bfield[..., 2], cmap='jet', vmin=vmin, vmax=vmax)
        # axs[2].set_title('Field Z [mT]')
        fig.tight_layout()
        
        plt.subplots_adjust(hspace=0, wspace=0)

    def visualize_encoding_functions(self,
                                     B0_encoding: torch.Tensor,
                                     B1_encoding: torch.Tensor,
                                     name: Optional[str] = 'encoding.gif'):
        """
        Visualizes the encoding functions

        Parameters:
        -----------
        B0_encoding : torch.Tensor
            Weighting from precession with shape (n_read, n_shots, *im_size)
        B1_encoding : torch.Tensor
            Receive coil weighting with shape (n_coils, n_shots, *im_size)
        """
        
        # B0 encoding visuals
        import matplotlib.animation as animation
        import matplotlib.pyplot as plt
        b0_phz = torch.rad2deg(torch.angle(B0_encoding)).cpu()
        b0_mag = torch.abs(B0_encoding).cpu()
        b1_mag = torch.abs(B1_encoding).cpu()
        vmax_b1 = (torch.median(b1_mag) + 3 * torch.std(b1_mag)).item()
        vmax_b0 = (torch.median(b0_mag) + 3 * torch.std(b0_mag)).item()
        coil = 7
        # times = torch.arange(0, B0_encoding.shape[0], 10)
        times = [40]
        ims = []
        if len(times) == 1:
            figsize=(9, 9)
        else:
            figsize=(14, 7)
        fig, axs = plt.subplots(nrows=3, ncols=3, figsize=figsize)
        for t in tqdm(times, 'Readout Plotting Loop'):
            im_batch = []
            for i in range(3):
                axs[0, i].set_title(f'Shot = {i}, t = {t}')
                for j in range(3):
                    axs[j, i].set_yticklabels([])
                    axs[j, i].set_xticklabels([])
                    axs[j, i].set_xticks([])
                    axs[j, i].set_yticks([])
                axs[0, 0].set_ylabel('Precession Magnitude')
                axs[1, 0].set_ylabel('Precession Phase')
                axs[2, 0].set_ylabel('Receive Coil Magnitude')
                im1 = axs[0, i].imshow(b0_mag[t, i], cmap='gray', vmin=0, vmax=vmax_b0)
                im2 = axs[1, i].imshow(b0_phz[t, i], cmap='jet', vmin=-180, vmax=180)
                im3 = axs[2, i].imshow(b1_mag[coil, i], cmap='gray', vmin=0, vmax=vmax_b1)
                im_batch += [im1, im2, im3]
            fig.tight_layout()
            ims.append(im_batch)

        
        plt.subplots_adjust(hspace=0, wspace=0)
        return

        # Add colorbars
        height = 0.28
        width = 0.02
        ofs = 0.01
        y1 = 0.18 - height/2 - ofs
        y2 = 0.5 - height/2 - ofs
        y3 = 1-0.18 - height/2 - ofs
        x = 0.74
        cb_ax = fig.add_axes([x,y1,width,height])
        fig.colorbar(im1,orientation='vertical',cax=cb_ax)
        cb_ax = fig.add_axes([x,y2,width,height])
        fig.colorbar(im2,orientation='vertical',cax=cb_ax)
        cb_ax = fig.add_axes([x,y3,width,height])
        fig.colorbar(im3,orientation='vertical',cax=cb_ax)

        print('Saving as .gif ...')
        ani = animation.ArtistAnimation(fig, ims, interval=20, blit=True,
                                repeat_delay=500)
        writer = animation.PillowWriter(fps=15,
                                metadata=dict(artist='Me'),
                                bitrate=1800,
                                )
        ani.save(name, writer=writer)
        


