import torch

from tqdm import tqdm
from typing import Optional
from dataclasses import dataclass
from einops import rearrange, einsum
from wearable_mri.utils import rotation_matrix

MU0 = 4e-7 * torch.pi # T*m/A

@dataclass
class wire_params:
        rho: Optional[float] = 1.62e-8 # Ohm m
        wire_diameter: Optional[float] = 1e-2 # m

class loop_coil:

    def __init__(self,
                 normal: torch.Tensor,
                 center: torch.Tensor,
                 diameter: Optional[float] = 0.1,
                 current: Optional[float] = 1,
                 nturns: Optional[int] = 100,
                 wparams: Optional[wire_params] = wire_params()):
        """
        Parameters:
        -----------
        normal : torch.Tensor
            Normal vector to the coil with shape (3,)
        center : torch.Tensor
            Center of the coil with shape (3,) [m]
        diameter : float
            diameter of the coil loop [m]
        current : float
            Current in the coil [A]
        nturns : int
            Number of turns in the coil
        wparams : wire_params
            Wire parameters
        """

        self.normal = normal
        self.center = center
        self.diameter = diameter
        self.current = current
        self.nturns = nturns
        self.wparams = wparams
    
    def show_info(self):
        """
        Prints the resistance, inductance, power, and Bfield
        """

        resistance = self.calc_resistance()
        inductance = self.calc_inductance()
        power = self.calc_power()
        B = self.calc_field_center()

        print(f"Resistance: {resistance*1e3:.3f} mOhms")
        print(f"Inductance: {inductance*1e3:.3f} mH")
        print(f"Power: {power:.3f} W")
        print(f"Bfield: {B*1e3:.3f} mT")

    def calc_resistance(self) -> float:
        """
        Calculates the resistance of the coil in ohms

        Returns:
        --------
        resistance : float
            Resistance of the coil [ohms]
        """

        # Coil parameters
        rho = self.wparams.rho
        diameter = self.diameter

        # Resistance
        A = torch.pi * (self.wparams.wire_diameter / 2) ** 2
        L = self.nturns * torch.pi * diameter
        resistance = rho * L / A

        return resistance
    
    def calc_inductance(self) -> float:
        """
        Calculates the inductance of the coil in henries

        Returns:
        --------
        inductance : float
            Inductance of the coil [H]
        """

        # Coil parameters
        diameter = self.diameter
        wire_diameter = self.wparams.wire_diameter

        l = self.nturns * wire_diameter
        A = torch.pi * (diameter / 2) ** 2
        inductance = MU0 * A * self.nturns ** 2 / l

        return inductance
    
    def calc_power(self) -> float:
        """
        Calculates the power dissipated in the coil

        Returns:
        --------
        power : float
            Power dissipated in the coil [W]
        """

        # Coil parameters
        current = self.current
        resistance = self.calc_resistance()

        # Power
        power = current ** 2 * resistance

        return power
    
    def calc_field_center(self) -> float:
        """
        Calculates the magnetic field at the center of coil

        Returns:
        --------
        B : float
            Field at center [T]
        """

        # Coil parameters
        current = self.current
        diameter = self.diameter

        # Magnetic field
        B = MU0 * current * self.nturns / (diameter)

        return B

    def calc_field(self,
                   pts: torch.Tensor) -> torch.Tensor:
        """
        Simulates the B0 field from a coil

        Parameters:
        -----------
        cparams : loop_coil
            Loop coil
        pts : torch.Tensor
            points to compute B field at with shape (..., 3)
        
        Returns:
        --------
        bfield : torch.Tensor
            magnetic field [T] at pts with shape (..., 3)
        """

        # Coil parameters
        normal = self.normal.type(torch.float32)
        normal = normal/torch.linalg.norm(normal)
        center = self.center.type(torch.float32)
        diameter = self.diameter
        current = self.current
        nturns = self.nturns

        # Sim params
        ntheta = 1000
        voxel_batch = 2 ** 10

        # GPU stuff
        torch_dev = pts.device
        normal = normal.to(torch_dev)
        center = center.to(torch_dev)

        # points of interest
        og_shape = pts.shape[:-1]
        pts = pts.reshape(-1, 3).type(torch.float32)
        N = len(pts)

        # Find orthogonal vector
        ortho = torch.randn(3, device=torch_dev, dtype=torch.float32)
        ortho = ortho - (ortho @ normal)*normal
        ortho /= torch.linalg.norm(ortho)

        # Precompute rotation terms for loop calculations
        thetas = torch.linspace(0, 2*torch.pi, ntheta, device=torch_dev, dtype=torch.float32)
        rots = rotation_matrix(normal, thetas) # 1000 3 3
        ls = einsum(ortho * diameter / 2, rots, 'three, N three three2 -> N three2')
        dls = torch.diff(ls, dim=0)
        ls = ls[:-1]
        
        # Sum magnetic field contributions along ring
        bfield = torch.zeros((N, 3), device=torch_dev, dtype=torch.float32)
        for n1 in tqdm(range(0, N, voxel_batch), 'Computing B0 Fields'):
            n2 = min(n1 + voxel_batch, N)

            # Distance from points to coil segments
            rs = pts[None, n1:n2, :] - (ls[:, None, :] + center[None, None]) # T P 3
            nrms = torch.linalg.norm(rs, axis=-1)[..., None]
            r_hats = rs/nrms

            # Biot Savart law
            bsl = torch.cross(dls[:, None, :], r_hats) / (nrms ** 2)
            bfield[n1:n2] = MU0 * current * nturns * torch.sum(bsl, axis=0) / (4*torch.pi)
        
        # Reshape
        bfield = bfield.reshape(og_shape + (3,))

        return bfield