import torch
import sigpy as sp

from wearable_mri.utils import torch_to_np, np_to_torch
from sigpy.linop import FiniteDifference

def finite_difference(x: torch.Tensor) -> torch.Tensor:
    """
    Compute finite difference

    Parameters:
    -----------
    x : torch.Tensor
        input tensor with shape (*im_size)

    Returns:
    --------
    dx : torch.Tensor
        finite difference of x with shape (2, *im_size)
    """
    x_sp = torch_to_np(x)
    dev = sp.get_device(x_sp)
    with dev:
        dx = np_to_torch(FiniteDifference(x_sp.shape) * x_sp)
    return dx

def finite_difference_adjoint(dx: torch.Tensor) -> torch.Tensor:
    """
    Compute adoint of finite difference

    Parameters:
    -----------
    dx : torch.Tensor
        finite difference of x with shape (2, *im_size)

    Returns:
    --------
    x : torch.Tensor
        adjoint with shape (*im_size)
    """
    dx_sp = torch_to_np(dx)
    dev = sp.get_device(dx_sp)
    with dev:
        x = np_to_torch(FiniteDifference(dx_sp.shape[1:]).H * dx_sp)
    return x
