import torch
import numpy as np
import cupy as cp

from typing import Optional

def normalize(shifted, target, ofs=True, mag=False):
    """
    Assumes the following scaling/shifting offset:

    shifted = a * target + b

    solves for a, b and returns the corrected data

    Parameters:
    -----------
    shifted : np.ndarray
        data to be corrected
    target : np.ndarray
        reference data
    ofs : bool
        include b offset in the correction
    mag : bool
        use magnitude of data for correction
    
    Returns:
    --------
    np.ndarray
        corrected data
    """
    is_torch = torch.is_tensor(shifted)
    target = torch_to_np(target)
    shifted = torch_to_np(shifted)
    try:
        if mag:
            col1 = np.abs(shifted).flatten()
            y = np.abs(target).flatten()
        else:
            col1 = shifted.flatten()
            y = target.flatten()

        if ofs:
            col2 = col1 * 0 + 1
            A = np.array([col1, col2]).T
            a, b = np.linalg.lstsq(A, y, rcond=None)[0]
        else:
            b = 0
            a = np.linalg.lstsq(np.array([col1]).T, y, rcond=None)[0]

        out = a * shifted + b
    except:
        print('Normalize Failed')
        out = shifted
    
    if is_torch:
        out = np_to_torch(out)
    return out


def rotation_matrix(axis: torch.Tensor, 
                    theta: torch.Tensor) -> torch.Tensor:
    """
    Computes rotation matrices for a given axis and angle

    Parameters:
    -----------
    axis : torch.Tensor
        axis of rotation with shape (..., 3)
    theta : torch.Tensor
        angle of rotation in radians with shape (...)
    
    Returns:
    --------
    R : torch.Tensor
        rotation matrix with shape (..., 3, 3)
    """
    
    dev = axis.device
    axis = axis / torch.linalg.norm(axis, dim=-1)
    a = torch.cos(theta / 2.0)
    b = -axis[..., 0] * torch.sin(theta / 2.0)
    c = -axis[..., 1] * torch.sin(theta / 2.0)
    d = -axis[..., 2] * torch.sin(theta / 2.0)
    R = torch.zeros((*theta.shape, 3, 3), device=dev, dtype=torch.float32)
    R[..., 0, 0] = a * a + b * b - c * c - d * d
    R[..., 0, 1] = 2 * (b * c - a * d)
    R[..., 0, 2] = 2 * (b * d + a * c)
    R[..., 1, 0] = 2 * (b * c + a * d)
    R[..., 1, 1] = a * a + c * c - b * b - d * d
    R[..., 1, 2] = 2 * (c * d - a * b)
    R[..., 2, 0] = 2 * (b * d - a * c)
    R[..., 2, 1] = 2 * (c * d + a * b)
    R[..., 2, 2] = a * a + d * d - b * b - c * c
    return R

def gen_grd(im_size: tuple, 
            fovs: Optional[tuple] = None) -> torch.Tensor:
    """
    Generates a grid of points given image size and FOVs

    Parameters:
    -----------
    im_size : tuple
        image dimensions
    fovs : tuple
        field of views, same size as im_size
    
    Returns:
    --------
    grd : torch.Tensor
        grid of points with shape (*im_size, len(im_size))
    """
    if fovs is None:
        fovs = (1,) * len(im_size)
    lins = [
        fovs[i] * torch.arange(-(im_size[i]//2), im_size[i]//2) / (im_size[i]) 
        for i in range(len(im_size))
        ]
    grds = torch.meshgrid(*lins, indexing='ij')
    grd = torch.cat(
        [g[..., None] for g in grds], dim=-1)
        
    return grd

def torch_to_np(*args):
    """
    Converts torch tensors to numpy arrays,
    preserving device and dtype (uses CUPY)

    Parameters:
    -----------
    args : tuple
        torch tensors to convert
    
    Returns:
    --------
    ret_args : tuple
        numpy arrays
    """
    ret_args = []
    for arg in args:
        if isinstance(arg, torch.Tensor):
            if arg.is_cuda:
                with cp.cuda.Device(arg.get_device()):
                    ret_args.append(cp.asarray(arg))
            else:
                ret_args.append(arg.numpy())
        elif isinstance(arg, np.ndarray) or isinstance(arg, cp.ndarray):
            ret_args.append(arg)
        else:
            ret_args.append(None)
    
    if len(ret_args) == 1:
        ret_args = ret_args[0]

    return ret_args

def np_to_torch(*args):
    """
    Converts numpy arrays to torch tensors,
    preserving device and dtype (uses CUPY)

    Parameters:
    -----------
    args : tuple
        numpy arrays to convert
    
    Returns:
    --------
    ret_args : tuple
        torch tensors
    """
    ret_args = []
    for arg in args:
        if isinstance(arg, np.ndarray):
            ret_args.append(torch.as_tensor(arg))
        elif isinstance(arg, cp.ndarray):
            ret_args.append(torch.as_tensor(arg, device=torch.device(int(arg.device))))
        elif isinstance(arg, torch.Tensor):
            ret_args.append(arg)
        else:
            ret_args.append(None)

    if len(ret_args) == 1:
        ret_args = ret_args[0]

    return ret_args
