import os
os.environ['CUDA_DEVICE_ORDER'] = 'PCI_BUS_ID'
import matplotlib
matplotlib.use('WebAgg')
import matplotlib.pyplot as plt

import torch
import numpy as np

from wearable_mri.utils import gen_grd, np_to_torch
from wearable_mri.encoding import encoding_linop
from wearable_mri.coil_geometries import (
    surface_coils_2d, 
    pre_polarizing_coils, 
    visualize_coils)

# Set random seeds
np.random.seed(0)
torch.manual_seed(0)

# Torch dev
device_idx = 5
torch_dev = torch.device(device_idx)
# torch_dev = torch.device('cpu')

# Params
show_encoding = False
n_shots = 200
K = 8 # coils
encode_current_max = 1
polarize_current_max = 10
dt = 4e-6
t_read = 2e-3
shot_batch_size = n_shots//50
coil_batch_size = 1
voxl_batch_size = 2 ** 10

# Phantom and sizes
phantom = np.load('data/brain_phantom.npy').astype(np.complex64)
phantom /= phantom.max()
diam_head = 0.22
im_size = phantom.shape

# points of interest
fovs = (diam_head, diam_head)
pts = gen_grd(im_size, fovs).to(torch_dev)
pts = torch.cat((pts, pts[..., :1] * 0), dim=-1).type(torch.float32)

# Build coil responses
coils_responses, lcs_enc = surface_coils_2d(n_coils=K,
                                   pts=pts,
                                   r_circle=diam_head / 2,
                                   r_coil=diam_head / 3,
                                   nturns=100)

# Polarizing field
polarizing_responses, lcs_pol = pre_polarizing_coils(pts, 
                                                     spacing=diam_head, 
                                                     r_coil=diam_head / 2, 
                                                     nturns=1000)
currents_ss = torch.tensor([1, 1]).to(torch_dev).type(torch.float32)
currents_ss *= polarize_current_max # Amps!!!
b0_ss = torch.sum(polarizing_responses * currents_ss[:, None, None, None], dim=0)

# Sequence selection
currents_imaging = 2 * torch.rand((n_shots, K), device=torch_dev, dtype=torch.float32) - 1
currents_imaging *= encode_current_max # Amps!!!

# Build encoding model 
times = torch.arange(0, t_read, dt).to(torch_dev).type(torch.float32)
A = encoding_linop(b0_steady_state=b0_ss, 
                   coil_responses=coils_responses, 
                   currents_imaging=currents_imaging,
                   times=times, 
                   shot_batch_size=shot_batch_size,
                   coil_batch_size=coil_batch_size, 
                   voxl_batch_size=voxl_batch_size,
                   verbose=False)

if show_encoding:
    probe = currents_imaging[0]
    A.visualize_field(probe)
    b0_enc, b1_enc = A.get_encoding_functions(shot_slice=slice(0,3))
    A.visualize_encoding_functions(b0_enc, b1_enc)
    plt.show()

# Simulate data
m = np_to_torch(phantom).to(torch_dev)
b = A.forward(m)

# Save imaging data
print(f'Saving acquired data ...', end='')
np.save('data/fake_data.npy', b.cpu().numpy())
print(' Done.')

# Compute AHA matrix (this can take some time)
A.calc_AHA_matrix(currents_imaging)

# Save Forward model
torch.save(A, 'data/encoding_model.pth')
