import os
os.environ['CUDA_DEVICE_ORDER'] = 'PCI_BUS_ID'
import matplotlib
matplotlib.use('WebAgg')
import matplotlib.pyplot as plt

import torch
import numpy as np

from wearable_mri.utils import np_to_torch
from wearable_mri.regularize.wavelet import wavelet
from wearable_mri.regularize.denoise import load_model
from wearable_mri.algs import (
    conjugate_gradient, 
    gradient_descent, 
    power_method_operator, 
    ADMM, 
    FISTA)
from wearable_mri.regularize.tv import (
    finite_difference, 
    finite_difference_adjoint)

# Set random seeds
np.random.seed(0)
torch.manual_seed(0)

# Torch dev
device_idx = 5
torch_dev = torch.device(device_idx)

# Recon params
sigma = 2e-3
n_params = 9
# rhos = [-1] * n_params
rhos = [1e-4] * n_params
# rhos = [5e-4] * n_params
# rhos = np.logspace(-4, -3, n_params)
lamda_l1s = np.logspace(-5, -4, n_params)
# lamda_l1s = np.logspace(-2, -1, n_params)
# lamda_l1s = [-1] * n_params
# lamda_l1s = [1e-2] * n_params
# lamda_l2s = np.logspace(-4, -2.5, n_params)
lamda_l2s = [-1] * n_params
# recon_alg = 'CG'
recon_alg = 'ADMM_wav'
# recon_alg = 'ADMM_tv'
# recon_alg = 'ADMM_dncnn'
save_data_path = 'data/recons/sweep/'

# load model and data
print(f'Loading data and encoding model ...', end='')
b = np_to_torch(np.load('data/fake_data.npy')).to(torch_dev)
A = torch.load('data/encoding_model.pth', map_location=torch_dev)
print(' done.')
b_noisy = b + sigma * torch.randn_like(b)

# Normalize by maximum eigenvalue
AHb = A.adjoint(b_noisy)
AHA = lambda x : A.normal(x)
im_size = A.im_size
x0 = torch.randn(im_size, dtype=torch.complex64, device=torch_dev)
_, lamda_max = power_method_operator(AHA, x0, num_iter=15)
AHA = lambda x : A.normal(x) / lamda_max
AHb = AHb / lamda_max ** 0.5

# Recons
for rho, lamda_l1, lamda_l2 in zip(rhos, lamda_l1s, lamda_l2s):
    if recon_alg == 'CG':
        x = conjugate_gradient(AHA, AHb, 
                                    num_iters=500, 
                                    lamda_l2=lamda_l2, 
                                    verbose=True)
    elif recon_alg == 'GD':
        x = gradient_descent(AHA, AHb, 
                                    lr=1e0, 
                                    num_iters=1000, 
                                    lamda_l2=0.0, 
                                    verbose=True)
    elif 'ADMM' in recon_alg:
        if 'tv' in recon_alg: 
            D = finite_difference
            DH = finite_difference_adjoint
            denoiser = None
        elif 'wav' in recon_alg:
            wav = wavelet(im_size)
            D = wav.forward
            DH = wav.adjoint
            denoiser = None
        elif 'dncnn' in recon_alg:
            denoiser = load_model(torch_dev)
            denoiser.eval()
            D = lambda x : x
            DH = lambda x : x
        x = ADMM(AHA=AHA, 
                D=D, 
                DH=DH, 
                AHb=AHb, 
                rho=rho, 
                lamda=lamda_l1, 
                num_iters=100,
                denoiser=denoiser)
    
    np.save(f'{save_data_path}{recon_alg}_{sigma:.3e}_{rho:.3e}_{lamda_l1:.3e}_{lamda_l2:.3e}', x.cpu().numpy())
