import os
import numpy as np

import matplotlib
matplotlib.use('WebAgg')
import matplotlib.pyplot as plt

from wearable_mri.utils import normalize
from skimage.metrics import peak_signal_noise_ratio 

matplotlib.rc('xtick', labelsize=20) 
matplotlib.rc('ytick', labelsize=20) 
font = {'family' : 'normal',
        'weight' : 'bold',
        'size'   : 22}
matplotlib.rc('font', **font)

def apply_inds(dct, inds):
    for key in dct:
        if type(dct[key]) is np.ndarray:
            dct[key] = dct[key][inds]
        else:
            dct[key] = np.array(dct[key])[inds]
    return dct

sweep_data_path = 'data/recons/sweep/'

# Pick param to sweep over and alg
param_keys = ['lamda_l2', 'lamda_l1', 'rho', 'sigma']
algs = ['CG', 'ADMM_tv', 'ADMM_wav', 'ADMM_dncnn']
sweep_param = 'lamda_l1'
alg = algs[2]
sigma_val = 2e-3

# Load phantom 
phantom = np.load('data/brain_phantom.npy').astype(np.complex64)

# load data
recons = []
params = {}
for key in param_keys:
    params[key] = []
for filename in os.listdir(sweep_data_path):
    if filename.startswith(alg):
        filename_no_ext = filename[:-4]
        splt = filename_no_ext.split('_')
        sigma, rho, lamda_l1, lamda_l2 = splt[-4:]
        params['sigma'].append(float(sigma))
        params['rho'].append(float(rho))
        params['lamda_l1'].append(float(lamda_l1))
        params['lamda_l2'].append(float(lamda_l2))
        recon = np.load(sweep_data_path + filename)
        recon = normalize(recon, phantom, ofs=True, mag=False)
        recons.append(recon)

# Sort by one parameter
inds = np.argwhere(np.isclose(params['sigma'], sigma_val)).flatten()
params = apply_inds(params, inds)
recons = np.array(recons)[inds]
sweep_vals = params[sweep_param]
inds = np.argsort(sweep_vals).flatten()
sweep_vals = np.array(sweep_vals)[inds]
params = apply_inds(params, inds)
recons = np.array(recons)[inds]

plt.figure(figsize=(14, 7))
for i, recon in enumerate(recons):
    plt.subplot(2, len(recons), i + 1)
    plt.imshow(np.abs(recon), cmap='gray')
    plt.axis('off')
    plt.subplot(2, len(recons), i + 1 + len(recons))
    plt.imshow(np.abs(recon - phantom), cmap='gray')
    plt.axis('off')
plt.tight_layout()
plt.subplots_adjust(hspace=0, wspace=0)

# Plot PSNR/RMSE vs param
nrsmes = []
psnrs = []
for i, recon in enumerate(recons):
    nrsmes.append(100 * np.linalg.norm(recon - phantom) / np.linalg.norm(phantom))
    data_range = np.max(np.abs(phantom)) - np.min(np.abs(phantom))
    psnr = peak_signal_noise_ratio(np.abs(phantom), np.abs(recon), data_range=data_range)
    psnrs.append(psnr)
fig, ax1 = plt.subplots(figsize=(14, 7))
ax1.set_xlabel(sweep_param)
ax1.semilogx(sweep_vals, nrsmes, color='r', linewidth=10)
ax1.set_ylabel('NRMSE (%)', color='r')
ax2 = ax1.twinx()
ax2.semilogx(sweep_vals, psnrs, color='b', linewidth=10)
ax2.set_ylabel('PSNR (dB)', color='b')
plt.show()

